/* liomux0.c
 * Linux
 * I/O Multiplexing (select) IPv4, IPv6, TCP Server, TCP client ohne after
 * C11 solution (Fortran style, like file descriptor open())
 * uses Array of Connection objects
 * uses Array element index
 * uses no class hierarchy (typ switch)
 * uses no business logic plugin
 *
 * Copyright 2022 Andre Adrian
 * See file license.txt for the 3-Clause BSD License for this source code
 *
 * 17may2022 adr: first published version
 * 29may2022 adr: use getnameinfo() instead of inet_ntop()
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <libgen.h>     // basename()

// Linux
#include <errno.h>
#include <unistd.h>
#include <netdb.h>
#include <sys/ioctl.h>
#include <sys/select.h>
#include <sys/time.h>
#include <arpa/inet.h>

#define FUNCTION __func__

enum {
  BUFSIZE = 1460,   // Ethernet packet size minus IPv4 TCP header size
  CONNMAX = 10,     // maximum number of Connection objects
  FDMAX = 64,       // maximum number of open file descriptors
  CONN_SERVER = 1,  // TCP server connection object label
  CONN_CLIENT = 2,  // TCP client connection object label
};

/* ******************************************************** */
// Connection object

typedef int CONN_ID;

typedef struct {
  fd_set fds;       // read file descriptor set
  int sockfd;       // network port for client, listen port for server
  int isConnecting; // non-blocking connect started, but no finished
  int typ;          // tells client or server object
} CONN;

static CONN conn[CONNMAX];  // Connection objects array

CONN_ID conn_factory() {
  for (CONN_ID id = 0; id < CONNMAX; ++id) {
    if (0 == conn[id].typ) {
      return id; // found a free entry
    }
  }
  assert(0 && "conn array full");
  return CONNMAX;
}

void client_open(CONN_ID id, const char* port, const char* hostname) {
  assert(id >= 0 && id < CONNMAX);
  assert(port != NULL);
  assert(hostname != NULL);
  printf("II %s: port=%s hostname=%s\n", FUNCTION, port, hostname);

  FD_ZERO(&conn[id].fds);
  conn[id].typ = CONN_CLIENT;

  // Try to connect
  struct addrinfo hints;
  memset(&hints, 0, sizeof hints);
  hints.ai_family = AF_UNSPEC;
  hints.ai_socktype = SOCK_STREAM;

  struct addrinfo* res;
  int rv = getaddrinfo(hostname, port, &hints, &res);
  if (rv != 0) {
    fprintf(stderr, "EE %s getaddrinfo: %s\n", FUNCTION, gai_strerror(rv));
    exit(EXIT_FAILURE);
  }

  // loop through all the results and connect to the first we can
  struct addrinfo* p;
  for (p = res; p != NULL; p = p->ai_next) {
    conn[id].sockfd = socket(p->ai_family, p->ai_socktype, p->ai_protocol);
    if (-1 == conn[id].sockfd) {
      perror("WW socket");
      continue;
    }

    // UNPv2 ch. 15.4, 16.2 non-blocking connect
    int val = 1;
    rv = ioctl(conn[id].sockfd, FIONBIO, &val);
    if (rv != 0) {
      perror("WW ioctl FIONBIO ON");
      close(conn[id].sockfd);
      continue;
    }

    rv = connect(conn[id].sockfd, p->ai_addr, p->ai_addrlen);
    if (rv != 0) {
      if (EINPROGRESS == errno) {
        conn[id].isConnecting = 1;
      } else {
        perror("WW connect");
        close(conn[id].sockfd);
        continue;
      }
    }

    break;  // exit loop after socket and connect were successful
  }

  assert(p != NULL && "connect try");

  char dst[NI_MAXHOST];
  rv = getnameinfo(p->ai_addr, p->ai_addrlen, dst, sizeof dst, NULL, 0, 0);
  assert(0 == rv && "getnameinfo");

  freeaddrinfo(res);

  printf("II %s: connect try to %s (%s) port %s socket %d\n", FUNCTION,
         hostname, dst, port, conn[id].sockfd);
}

void client_handle(CONN_ID id, int fd) {
  assert(id >= 0 && id < CONNMAX);
  assert(fd >= 0 && fd < FDMAX);
  int cb_client_read(CONN_ID id, int fd);
  int rv = cb_client_read(id, fd);
  if (rv < 1) {
    int optval = 0;
    socklen_t optlen = sizeof optval;
    int rv = getsockopt(conn[id].sockfd, SOL_SOCKET, SO_ERROR, &optval, &optlen);
    assert(0 == rv && "getsockopt SOL_SOCKET SO_ERROR");
    fprintf(stderr, "EE %s: connect fail to socket %d rv %d: %s\n",
            FUNCTION, conn[id].sockfd, rv, strerror(optval));
    assert(0);
  }
}

void conn_add_fd(CONN_ID id, int fd) {
  assert(id >= 0 && id < CONNMAX);
  assert(fd >= 0 && fd < FDMAX);
  FD_SET(fd, &conn[id].fds);
}

void client_connect(CONN_ID id) {
  assert(id >= 0 && id < CONNMAX);
  conn[id].isConnecting = 0;
  int optval = 0;
  socklen_t optlen = sizeof optval;
  int rv = getsockopt(conn[id].sockfd, SOL_SOCKET, SO_ERROR, &optval, &optlen);
  assert(0 == rv && "getsockopt SOL_SOCKET SO_ERROR");
  if (0 == optval) {
    FD_SET(conn[id].sockfd, &conn[id].fds);
    printf("II %s: connect success to socket %d\n", FUNCTION,
           conn[id].sockfd);
  } else {
    fprintf(stderr, "EE %s: connect fail to socket %d: %s\n", FUNCTION,
            conn[id].sockfd, strerror(optval));
    assert(0);
  }
}

void server_open(CONN_ID id, const char* port, const char* hostname) {
  assert(id >= 0 && id < CONNMAX);
  assert(port != NULL);
  // no assert hostname
  printf("II %s: port=%s hostname=%s\n", FUNCTION, port, hostname);

  FD_ZERO(&conn[id].fds);
  conn[id].typ = CONN_SERVER;

  struct addrinfo hints;
  memset(&hints, 0, sizeof hints);
  hints.ai_family = AF_UNSPEC;
  hints.ai_socktype = SOCK_STREAM;
  hints.ai_flags = AI_PASSIVE;

  struct addrinfo* res;
  int rv = getaddrinfo(hostname, port, &hints, &res);
  if (rv != 0) {
    fprintf(stderr, "EE %s getaddrinfo: %s\n", FUNCTION, gai_strerror(rv));
    exit(EXIT_FAILURE);
  }

  struct addrinfo* p;
  for(p = res; p != NULL; p = p->ai_next) {
    conn[id].sockfd = socket(p->ai_family, p->ai_socktype, p->ai_protocol);
    if (-1 == conn[id].sockfd) {
      perror("WW socket");
      continue;
    }

    int yes = 1;
    rv = setsockopt(conn[id].sockfd, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(yes));
    if (rv != 0) {
      perror("WW setsockopt SO_REUSEADDR");
      close(conn[id].sockfd);
      continue;
    }

    rv = bind(conn[id].sockfd, p->ai_addr, p->ai_addrlen);
    if (rv != 0) {
      perror("WW bind");
      close(conn[id].sockfd);
      continue;
    }

    break;  // exit loop after socket and bind were successful
  }

  freeaddrinfo(res);

  assert(p != NULL && "bind");

  rv = listen(conn[id].sockfd, 10);
  assert(0 == rv && "listen");

  // add the listener to the fd_set
  FD_SET(conn[id].sockfd, &conn[id].fds);

  printf("II %s: listen on socket %d\n", FUNCTION, conn[id].sockfd);
}

void server_handle(CONN_ID id, int fd) {
  assert(id >= 0 && id < CONNMAX);
  assert(fd >= 0 && fd < FDMAX);
  if (fd == conn[id].sockfd) {
    // handle new connections
    struct sockaddr_storage remoteaddr; // client address
    socklen_t addrlen = sizeof remoteaddr;
    // newly accept()ed socket descriptor
    int newfd = accept(conn[id].sockfd, (struct sockaddr*)&remoteaddr, &addrlen);
    if (-1 == newfd) {
      perror("WW accept");
    } else {
      FD_SET(newfd, &conn[id].fds);
      char dst[NI_MAXHOST];
      int rv = getnameinfo((struct sockaddr*)&remoteaddr, sizeof remoteaddr, dst,
                           sizeof dst, NULL, 0, 0);
      assert(0 == rv && "getnameinfo");
      printf("II %s: new connection %s on socket %d\n", FUNCTION, dst, newfd);
    }
  } else {
    int cb_server_read(CONN_ID id, int fd);
    int rv = cb_server_read(id, fd);
    if (rv < 1) {
      printf("II %s: connection closed on socket %d\n", FUNCTION, fd);
      close(fd);
      FD_CLR(fd, &conn[id].fds); // remove from fd_set
    }
  }
}

void conn_event_loop() {
  for(;;) {
    // virtualization pattern: join all read fds into one
    fd_set read_fds = conn[0].fds;
    for (int iconn = 1; iconn < CONNMAX; ++iconn) {
      for (int fd = 0; fd < FDMAX; ++fd) {
        if (FD_ISSET(fd, &conn[iconn].fds)) {
          FD_SET(fd, &read_fds);
        }
      }
    }
    // virtualization pattern: join all connect pending into one
    fd_set write_fds;
    FD_ZERO(&write_fds);
    for (int iconn = 0; iconn < CONNMAX; ++iconn) {
      if (conn[iconn].isConnecting) {
        FD_SET(conn[iconn].sockfd, &write_fds);
      }
    }

    int rv = select(FDMAX, &read_fds, &write_fds, NULL, NULL);
    if (-1 == rv && EINTR != errno) {
      perror("EE select");
      exit(EXIT_FAILURE);
    }

    if (rv > 0) {
      // looking for data to read available
      for (int fd = 0; fd < FDMAX; ++fd) {
        if (FD_ISSET(fd, &read_fds)) {
          for (int iconn = 0; iconn < CONNMAX; ++iconn) {
            if (FD_ISSET(fd, &conn[iconn].fds)) {
              switch (conn[iconn].typ) {
              case CONN_CLIENT:
                client_handle(iconn, fd);
                break;
              case CONN_SERVER:
                server_handle(iconn, fd);
                break;
              }
            }
          }
        }
      }
      // looking for connect pending success or fail
      for (int iconn = 0; iconn < CONNMAX; ++iconn) {
        if (FD_ISSET(conn[iconn].sockfd, &write_fds)) {
          client_connect(iconn);
        }
      }
    }
  }
}

/* ******************************************************** */
// Business logic

/** @brief callback client read available
 * @retval >0 okay
 * @retval <=0 close or error
 */
int cb_client_read(CONN_ID id, int fd) {
  assert(id >= 0 && id < CONNMAX);
  assert(fd >= 0 && fd < FDMAX);
  // Compute where to write data. If we're stdin (0),
  // we'll write to the sockfd. If we're the sockfd, we'll
  // write to stdout (1).
  int outfd = (STDIN_FILENO == fd)? conn[id].sockfd: STDOUT_FILENO;

  // We use read() and write() in here since those work on
  // all fds, not just sockets. send() and recv() would
  // fail on stdin and stdout since they're not sockets.
  char buf[BUFSIZE];
  int readbytes = read(fd, buf, sizeof buf);
  if (readbytes > 0) {
    // Write all data out
    int writebytes = write(outfd, buf, readbytes);
    assert(writebytes == readbytes && "write");
  }
  return readbytes;
}

/** @brief callback server read available
 * @retval >0 okay
 * @retval <=0 close or error
 */
int cb_server_read(CONN_ID id, int fd) {
  assert(id >= 0 && id < CONNMAX);
  assert(fd >= 0 && fd < FDMAX);
  char buf[BUFSIZE];    // buffer for client data
  int readbytes = read(fd, buf, sizeof buf);
  if (readbytes > 0) {
    // we got some data from a client
    for (int i = 0; i < FDMAX; ++i) {
      // send to everyone!
      if (FD_ISSET(i, &conn[id].fds)) {
        // except the listener and ourselves
        if (i != conn[id].sockfd && i != fd) {
          int writebytes = write(i, buf, readbytes);
          if (writebytes != readbytes) {
            perror("WW send");
          }
        }
      }
    }
  }
  return readbytes;
}

int main(int argc, char* argv[]) {
  if (argc < 3) {
    char* name = basename(argv[0]);
    fprintf(stderr,"usage server: %s s port [hostname]\n", name);
    fprintf(stderr,"usage client: %s c hostname port\n", name);
    fprintf(stderr,"example server IPv4: ./%s s 60000\n", name);
    fprintf(stderr,"example client IPv4: ./%s c 127.0.0.1 60000\n", name);
    fprintf(stderr,"example server IPv6: ./%s s 60000 ::1\n", name);
    fprintf(stderr,"example client IPv6: ./%s c ::1 60000\n", name);
    exit(EXIT_FAILURE);
  }

  switch(argv[1][0]) {
  case 'c': {
    CONN_ID id = conn_factory();
    client_open(id, argv[3], argv[2]);
    conn_add_fd(id, STDIN_FILENO);
  }
  break;
  case 's': {
    CONN_ID id = conn_factory();
    server_open(id, argv[2], argv[3]);
  }
  break;
  default:
    fprintf(stderr,"EE %s: unexpected argument %s\n", FUNCTION, argv[1]);
    exit(EXIT_FAILURE);
  }

  conn_event_loop();  // start inversion of control
  return 0;
}
