/* liomux3.c
 * Linux
 * I/O Multiplexing (select) IPv4, IPv6, TCP Server, TCP client
 * C11 solution (C framework/library style)
 * uses Array of Connection objects
 * uses Array element pointer
 * uses no class hierarchy (typ switch)
 * uses business logic plugin via callbacks
 *
 * Copyright 2022 Andre Adrian
 * See file license.txt for the 3-Clause BSD License for this source code
 *
 * 17may2022 adr: first published version
 * 29may2022 adr: use getnameinfo() instead of inet_ntop()
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <libgen.h>     // basename()
#include <time.h>

// Linux
#include <errno.h>
#include <unistd.h>
#include <netdb.h>
#include <sys/ioctl.h>
#include <sys/select.h>
#include <sys/time.h>
#include <arpa/inet.h>

#define FUNCTION __func__

enum {
  BUFSIZE = 1460,   // Ethernet packet size minus IPv4 TCP header size
  CONNMAX = 10,     // maximum number of Connection objects
  FDMAX = 64,       // maximum number of open file descriptors
  CONN_SERVER = 1,  // TCP server connection object label
  CONN_CLIENT = 2,  // TCP client connection object label
  TIMEOUT = 40,     // select() timeout in milli seconds
  STRMAX = 80,      // maximum length of C-String
  TIMERMAX = 10,    // maximum number of Timer objects
};

// Copies a string with security enhancements
void strcpy_s(char* dest, size_t n, const char* src) {
  strncpy(dest, src, n);
  dest[n - 1] = '\0';
}

/* ******************************************************** */
// Timer object

typedef void (*cb_timer_t)(void* obj);

typedef struct  {
  cb_timer_t cb_timer;  // cb_timer and obj are a closure
  void* obj;
  struct timespec ts;   // expire time
} TIMER;

static TIMER timer[TIMERMAX];  // Timer objects array

int after(int interval, cb_timer_t cb_timer, void* obj) {
  assert(interval >= 0);
  assert(cb_timer != NULL);
  // no assert obj

  int id;
  for (id = 0; id < TIMERMAX; ++id) {
    if (NULL == timer[id].cb_timer) {
      break;  // found a free entry
    }
  }
  assert (id < TIMERMAX && "timer array full");

  // convert interval in milliseconds to timespec
  struct timespec dts;
  dts.tv_nsec = (interval % 1000) * 1000000;
  dts.tv_sec = interval / 1000;
  struct timespec now;
  clock_gettime(CLOCK_MONOTONIC, &now);
  timer[id].cb_timer = cb_timer;
  timer[id].obj = obj;
  timer[id].ts.tv_nsec = (now.tv_nsec + dts.tv_nsec) % 1000000000;
  timer[id].ts.tv_sec = (now.tv_nsec + dts.tv_nsec) / 1000000000;
  timer[id].ts.tv_sec += (now.tv_sec + dts.tv_sec);
  /*
  printf("II %s now=%ld,%ld dt=%ld,%ld ts=%ld,%ld\n", FUNCTION,
    now.tv_sec, now.tv_nsec, dts.tv_sec, dts.tv_nsec,
    timer[i].ts.tv_sec, timer[i].ts.tv_nsec);
  */
  return id;
}

void timer_walk() {
  // looking for expired timers
  for (int i = 0; i < TIMERMAX; ++i) {
    if (timer[i].cb_timer != NULL) {
      struct timespec ts;
      clock_gettime(CLOCK_MONOTONIC, &ts);
      if ((ts.tv_sec > timer[i].ts.tv_sec) || (ts.tv_sec == timer[i].ts.tv_sec
          && ts.tv_nsec >= timer[i].ts.tv_nsec)) {
        TIMER tmp = timer[i];
        // erase array entry because called function can overwrite this entry
        memset(&timer[i], 0, sizeof timer[i]);
        assert(tmp.cb_timer != NULL);
        (*tmp.cb_timer)(tmp.obj);
      }
    }
  }
}

/* ******************************************************** */
// Connection object

struct CONN_ {
  fd_set fds;       // read file descriptor set
  int sockfd;       // network port for client, listen port for server
  int isConnecting; // non-blocking connect started, but no finished
  int typ;          // tells client or server object
  char hostname[STRMAX];
  char port[STRMAX];
  int (*cb_read)(struct CONN_* obj, int fd);   // Callback Pointer
};

typedef struct CONN_ CONN;
typedef int (*cb_read_t)(CONN* obj, int fd);  // Callback Pointer type

static CONN conn[CONNMAX];  // Connection objects array

CONN* conn_factory() {
  for (int i = 0; i < CONNMAX; ++i) {
    if (0 == conn[i].typ) {
      return &conn[i];  // found a free entry
    }
  }
  assert(0 && "conn array full");
  return NULL;
}

void client_open(CONN* obj, const char* port, const char* hostname) {
  assert(obj >= &conn[0] && obj < &conn[CONNMAX]);
  assert(port != NULL);
  assert(hostname != NULL);
  printf("II %s: port=%s hostname=%s\n", FUNCTION, port, hostname);

  FD_ZERO(&obj->fds);
  obj->typ = CONN_CLIENT;

  strcpy_s(obj->port, sizeof obj->port, port);
  strcpy_s(obj->hostname, sizeof obj->hostname, hostname);
  void client_open1(CONN* obj);
  client_open1(obj);
}

void client_open1(CONN* obj) {
  assert(obj >= &conn[0] && obj < &conn[CONNMAX]);
  struct addrinfo hints;
  memset(&hints, 0, sizeof hints);
  hints.ai_family = AF_UNSPEC;
  hints.ai_socktype = SOCK_STREAM;

  struct addrinfo* res;
  int rv = getaddrinfo(obj->hostname, obj->port, &hints, &res);
  if (rv != 0) {
    fprintf(stderr, "EE %s getaddrinfo: %s\n", FUNCTION, gai_strerror(rv));
    exit(EXIT_FAILURE);
  }

  // loop through all the results and connect to the first we can
  struct addrinfo* p;
  for (p = res; p != NULL; p = p->ai_next) {
    obj->sockfd = socket(p->ai_family, p->ai_socktype, p->ai_protocol);
    if (-1 == obj->sockfd) {
      perror("WW socket");
      continue;
    }

    // UNPv2 ch. 15.4, 16.2 non-blocking connect
    int val = 1;
    rv = ioctl(obj->sockfd, FIONBIO, &val);
    if (rv != 0) {
      perror("WW ioctl FIONBIO ON");
      close(obj->sockfd);
      continue;
    }

    rv = connect(obj->sockfd, p->ai_addr, p->ai_addrlen);
    if (rv != 0) {
      if (EINPROGRESS == errno) {
        obj->isConnecting = 1;
      } else {
        perror("WW connect");
        close(obj->sockfd);
        continue;
      }
    }

    break;  // exit loop after socket and connect were successful
  }

  assert(p != NULL && "connect try");

  char dst[NI_MAXHOST];
  rv = getnameinfo(p->ai_addr, p->ai_addrlen, dst, sizeof dst, NULL, 0, 0);
  assert(0 == rv && "getnameinfo");

  freeaddrinfo(res);

  // don't add sockfd to the fd_set, client_connect() will do

  printf("II %s: connect try to %s (%s) port %s socket %d\n", FUNCTION,
         obj->hostname, dst, obj->port, obj->sockfd);
}

void client_reopen(CONN* obj) {
  assert(obj >= &conn[0] && obj < &conn[CONNMAX]);
  close(obj->sockfd);
  FD_CLR(obj->sockfd, &obj->fds);   // remove network fd
  obj->sockfd = -1;
  after(5000, (cb_timer_t)client_open1, obj);
  // ugly bug with after(0, ...
}

void client_handle(CONN* obj, int fd) {
  assert(obj >= &conn[0] && obj < &conn[CONNMAX]);
  assert(fd >= 0 && fd < FDMAX);
  assert(obj->cb_read != NULL);
  int rv = (*obj->cb_read)(obj, fd);
  if (rv < 1) {
    int optval = 0;
    socklen_t optlen = sizeof optval;
    int rv = getsockopt(obj->sockfd, SOL_SOCKET, SO_ERROR, &optval, &optlen);
    assert(0 == rv && "getsockopt SOL_SOCKET SO_ERROR");
    fprintf(stderr, "WW %s: connect fail to %s port %s socket %d rv %d: %s\n",
            FUNCTION, obj->hostname, obj->port, obj->sockfd, rv, strerror(optval));
    client_reopen(obj);
  }
}

void conn_add_fd(CONN* obj, int fd) {
  assert(obj >= &conn[0] && obj < &conn[CONNMAX]);
  assert(fd >= 0 && fd < FDMAX);
  FD_SET(fd, &obj->fds);
}

void conn_add_cb(CONN* obj, cb_read_t cb_read) {
  assert(cb_read != NULL);
  obj->cb_read = cb_read;
}

void client_connect(CONN* obj) {
  assert(obj >= &conn[0] && obj < &conn[CONNMAX]);
  obj->isConnecting = 0;
  int optval = 0;
  socklen_t optlen = sizeof optval;
  int rv = getsockopt(obj->sockfd, SOL_SOCKET, SO_ERROR, &optval, &optlen);
  assert(0 == rv && "getsockopt SOL_SOCKET SO_ERROR");
  if (0 == optval) {
    FD_SET(obj->sockfd, &obj->fds);
    printf("II %s: connect success to %s port %s socket %d\n", FUNCTION,
           obj->hostname, obj->port, obj->sockfd);
  } else {
    fprintf(stderr, "WW %s: connect fail to %s port %s socket %d: %s\n", FUNCTION,
            obj->hostname, obj->port, obj->sockfd, strerror(optval));
    client_reopen(obj);
  }
}

void server_open(CONN* obj, const char* port, const char* hostname) {
  assert(obj >= &conn[0] && obj < &conn[CONNMAX]);
  assert(port != NULL);
  // no assert hostname
  printf("II %s: port=%s hostname=%s\n", FUNCTION, port, hostname);

  FD_ZERO(&obj->fds);
  obj->typ = CONN_SERVER;

  struct addrinfo hints;
  memset(&hints, 0, sizeof hints);
  hints.ai_family = AF_UNSPEC;
  hints.ai_socktype = SOCK_STREAM;
  hints.ai_flags = AI_PASSIVE;

  struct addrinfo* res;
  int rv = getaddrinfo(hostname, port, &hints, &res);
  if (rv != 0) {
    fprintf(stderr, "EE %s getaddrinfo: %s\n", FUNCTION, gai_strerror(rv));
    exit(EXIT_FAILURE);
  }

  struct addrinfo* p;
  for(p = res; p != NULL; p = p->ai_next) {
    obj->sockfd = socket(p->ai_family, p->ai_socktype, p->ai_protocol);
    if (-1 == obj->sockfd) {
      perror("WW socket");
      continue;
    }

    int yes = 1;
    rv = setsockopt(obj->sockfd, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(yes));
    if (rv != 0) {
      perror("WW setsockopt SO_REUSEADDR");
      close(obj->sockfd);
      continue;
    }

    rv = bind(obj->sockfd, p->ai_addr, p->ai_addrlen);
    if (rv != 0) {
      perror("WW bind");
      close(obj->sockfd);
      continue;
    }

    break;  // exit loop after socket and bind were successful
  }

  freeaddrinfo(res);

  assert(p != NULL && "bind");

  rv = listen(obj->sockfd, 10);
  assert(0 == rv && "listen");

  // add the listener to the master set
  FD_SET(obj->sockfd, &obj->fds);

  printf("II %s: listen on socket %d\n", FUNCTION, obj->sockfd);
}

void server_handle(CONN* obj, int fd) {
  assert(obj >= &conn[0] && obj < &conn[CONNMAX]);
  assert(fd >= 0 && fd < FDMAX);
  if (fd == obj->sockfd) {
    // handle new connections
    struct sockaddr_storage remoteaddr; // client address
    socklen_t addrlen = sizeof remoteaddr;
    // newly accept()ed socket descriptor
    int newfd = accept(obj->sockfd, (struct sockaddr*)&remoteaddr, &addrlen);
    if (-1 == newfd) {
      perror("WW accept");
    } else {
      FD_SET(newfd, &obj->fds);
      char dst[NI_MAXHOST];
      int rv = getnameinfo((struct sockaddr*)&remoteaddr, sizeof remoteaddr, dst,
                           sizeof dst, NULL, 0, 0);
      assert(0 == rv && "getnameinfo");
      printf("II %s: new connection %s on socket %d\n", FUNCTION, dst, newfd);
    }
  } else {
    assert(obj->cb_read != NULL);
    int rv = (*obj->cb_read)(obj, fd);
    if (rv < 1) {
      printf("II %s: connection closed on socket %d\n", FUNCTION, fd);
      close(fd);
      FD_CLR(fd, &obj->fds); // remove from fd_set
    }
  }
}

void conn_event_loop() {
  for(;;) {
    // virtualization pattern: join all read fds into one
    fd_set read_fds = conn[0].fds;
    for (int iconn = 1; iconn < CONNMAX; ++iconn) {
      for (int fd = 0; fd < FDMAX; ++fd) {
        if (FD_ISSET(fd, &conn[iconn].fds)) {
          FD_SET(fd, &read_fds);
        }
      }
    }
    // virtualization pattern: join all connect pending into one
    fd_set write_fds;
    FD_ZERO(&write_fds);
    for (int iconn = 0; iconn < CONNMAX; ++iconn) {
      if (conn[iconn].isConnecting) {
        FD_SET(conn[iconn].sockfd, &write_fds);
      }
    }

    struct timeval tv = {0, TIMEOUT * 1000};
    int rv = select(FDMAX, &read_fds, &write_fds, NULL, &tv);
    if (-1 == rv && EINTR != errno) {
      perror("EE select");
      exit(EXIT_FAILURE);
    }

    if (rv > 0) {
      // looking for data to read available
      for (int fd = 0; fd < FDMAX; ++fd) {
        if (FD_ISSET(fd, &read_fds)) {
          for (int iconn = 0; iconn < CONNMAX; ++iconn) {
            if (FD_ISSET(fd, &conn[iconn].fds)) {
              switch (conn[iconn].typ) {
              case CONN_CLIENT:
                client_handle(&conn[iconn], fd);
                break;
              case CONN_SERVER:
                server_handle(&conn[iconn], fd);
                break;
              }
            }
          }
        }
      }
      // looking for connect pending success or fail
      for (int iconn = 0; iconn < CONNMAX; ++iconn) {
        if (FD_ISSET(conn[iconn].sockfd, &write_fds)) {
          client_connect(&conn[iconn]);
        }
      }
    }
    timer_walk();
  }
}

/* ******************************************************** */
// Business logic

/** @brief callback client read available
 * @retval >0 okay
 * @retval <=0 close or error
 */
int cb_chat_client_read(CONN* obj, int fd) {
  assert(obj >= &conn[0] && obj < &conn[CONNMAX]);
  assert(fd >= 0 && fd < FDMAX);
  // Compute where to write data. If we're stdin (0),
  // we'll write to the sockfd. If we're the sockfd, we'll
  // write to stdout (1).
  int outfd = (STDIN_FILENO == fd)? obj->sockfd: STDOUT_FILENO;

  // We use read() and write() in here since those work on
  // all fds, not just sockets. send() and recv() would
  // fail on stdin and stdout since they're not sockets.
  char buf[BUFSIZE];
  int readbytes = read(fd, buf, sizeof buf);
  if (readbytes > 0) {
    // Write all data out
    int writebytes = write(outfd, buf, readbytes);
    assert(writebytes == readbytes && "write");
  }
  return readbytes;
}

/** @brief callback server read available
 * @retval >0 okay
 * @retval <=0 close or error
 */
int cb_chat_server_read(CONN* obj, int fd) {
  assert(obj >= &conn[0] && obj < &conn[CONNMAX]);
  assert(fd >= 0 && fd < FDMAX);
  char buf[BUFSIZE];    // buffer for client data
  int readbytes = read(fd, buf, sizeof buf);
  if (readbytes > 0) {
    // we got some data from a client
    for (int i = 0; i < FDMAX; ++i) {
      // send to everyone!
      if (FD_ISSET(i, &obj->fds)) {
        // except the listener and ourselves
        if (i != obj->sockfd && i != fd) {
          int writebytes = write(i, buf, readbytes);
          if (writebytes != readbytes) {
            perror("WW send");
          }
        }
      }
    }
  }
  return readbytes;
}

int main(int argc, char* argv[]) {
  if (argc < 3) {
    char* name = basename(argv[0]);
    fprintf(stderr,"usage server: %s s port [hostname]\n", name);
    fprintf(stderr,"usage client: %s c hostname port\n", name);
    fprintf(stderr,"example server IPv4: ./%s s 60000\n", name);
    fprintf(stderr,"example client IPv4: ./%s c 127.0.0.1 60000\n", name);
    fprintf(stderr,"example server IPv6: ./%s s 60000 ::1\n", name);
    fprintf(stderr,"example client IPv6: ./%s c ::1 60000\n", name);
    exit(EXIT_FAILURE);
  }

  switch(argv[1][0]) {
  case 'c': {
    CONN* obj = conn_factory();
    client_open(obj, argv[3], argv[2]);
    conn_add_cb(obj, cb_chat_client_read);
    conn_add_fd(obj, STDIN_FILENO);
  }
  break;
  case 's': {
    CONN* obj = conn_factory();
    server_open(obj, argv[2], argv[3]);
    conn_add_cb(obj, cb_chat_server_read);
  }
  break;
  default:
    fprintf(stderr,"EE %s: unexpected argument %s\n", FUNCTION, argv[1]);
    exit(EXIT_FAILURE);
  }

  conn_event_loop();  // start inversion of control
  return 0;
}
