/* wiomux4.cpp
 * MS-Windows
 * I/O Multiplexing (select) IPv4, IPv6, TCP Server, TCP client
 * C++14 solution (OO style)
 * uses Array of Connection objects
 * uses Object Oriented hidden pointer
 * uses no class hierarchy (typ switch)
 * uses no business logic plugin
 *
 * Copyright 2022 Andre Adrian
 * See file license.txt for the 3-Clause BSD License for this source code
 *
 * 22may2022 adr: first published version
 */

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <cassert>
#include <libgen.h>     // basename()
#include <time.h>

// MS-Windows
#include <conio.h>      // _kbhit()
#include <winsock2.h>
#include <ws2tcpip.h>   // getaddrinfo()

#define FUNCTION __PRETTY_FUNCTION__

enum {
  BUFSIZE = 1460,   // Ethernet packet size minus IPv4 TCP header size
  CONNMAX = 10,     // maximum number of Connection objects
  FDMAX = 256,      // maximum number of open file descriptors
  CONN_SERVER = 1,  // TCP server connection object label
  CONN_CLIENT = 2,  // TCP client connection object label
  TIMEOUT = 40,     // select() timeout in milli seconds
  STRMAX = 80,      // maximum length of C-String
  TIMERMAX = 10,    // maximum number of Timer objects
};

// get sockaddr, IPv4 or IPv6:
void* get_in_addr(struct sockaddr* sa) {
  assert(sa != nullptr);
  if (AF_INET == sa->sa_family) {
    return &(reinterpret_cast<struct sockaddr_in*>(sa)->sin_addr);
  } else {
    return &(reinterpret_cast<struct sockaddr_in6*>(sa)->sin6_addr);
  }
}

/* ******************************************************** */
// Timer object

using cb_timer_t = void (*)(void* obj);

class TIMER {
  cb_timer_t cb_timer;  // cb_timer and obj are a closure
  void* obj;
  struct timespec ts;   // expire time
public:
  static int after(int interval, cb_timer_t cb_timer, void* obj);
  static void walk();
};

static TIMER timer[TIMERMAX];  // Timer objects array

int TIMER::after(int interval, cb_timer_t cb_timer, void* obj) {
  assert(interval >= 0);
  assert(cb_timer != nullptr);
  // no assert obj

  int id;
  for (id = 0; id < TIMERMAX; ++id) {
    if (nullptr == timer[id].cb_timer) {
      break;  // found a free entry
    }
  }
  assert (id < TIMERMAX && "timer array full");

  // convert interval in milliseconds to timespec
  struct timespec dts;
  dts.tv_nsec = (interval % 1000) * 1000000;
  dts.tv_sec = interval / 1000;
  struct timespec now;
  clock_gettime(CLOCK_MONOTONIC, &now);
  timer[id].cb_timer = cb_timer;
  timer[id].obj = obj;
  timer[id].ts.tv_nsec = (now.tv_nsec + dts.tv_nsec) % 1000000000;
  timer[id].ts.tv_sec = (now.tv_nsec + dts.tv_nsec) / 1000000000;
  timer[id].ts.tv_sec += (now.tv_sec + dts.tv_sec);
  /*
  printf("II %s now=%ld,%ld dt=%ld,%ld ts=%ld,%ld\n", FUNCTION,
    now.tv_sec, now.tv_nsec, dts.tv_sec, dts.tv_nsec,
    timer[i].ts.tv_sec, timer[i].ts.tv_nsec);
  */
  return id;
}

void TIMER::walk() {
  // looking for expired timers
  for (int i = 0; i < TIMERMAX; ++i) {
    if (timer[i].cb_timer != nullptr) {
      struct timespec ts;
      clock_gettime(CLOCK_MONOTONIC, &ts);
      if ((ts.tv_sec > timer[i].ts.tv_sec) || (ts.tv_sec == timer[i].ts.tv_sec
          && ts.tv_nsec >= timer[i].ts.tv_nsec)) {
        TIMER tmp = timer[i];
        // erase array entry because called function can overwrite this entry
        memset(&timer[i], 0, sizeof timer[i]);
        assert(tmp.cb_timer != nullptr);
        (*tmp.cb_timer)(tmp.obj);
      }
    }
  }
}

/* ******************************************************** */
// Connection object

class CONN {
  fd_set fds;       // read file descriptor set
  SOCKET sockfd;    // network port for client, listen port for server
  int isConnecting; // non-blocking connect started, but no finished
  int typ;          // tells client or server object
  char hostname[STRMAX];
  char port[STRMAX];

public:
  CONN() : sockfd(-1), isConnecting(0), typ(0) {
    FD_ZERO(&fds);
  };
  static CONN* factory();   // Klassenmethode
  void client_open(const char* port, const char* hostname);
  void client_open1();
  void client_connect();
  void client_reopen();
  void client_handle(SOCKET fd);
  int cb_client_read(SOCKET fd);

  void server_open(const char* port, const char* hostname);
  void server_handle(SOCKET fd);
  int cb_server_read(SOCKET fd);

  void add_fd(SOCKET fd);
  static int get_sockfd(CONN* obj);
  static void event_loop();
};

static CONN conn[CONNMAX];  // Connection objects array

CONN* CONN::factory() {
  for (int i = 0; i < CONNMAX; ++i) {
    if (0 == conn[i].typ) {
      return &conn[i];  // found a free entry
    }
  }
  assert(0 && "conn array full");
  return nullptr;
}

int CONN::get_sockfd(CONN* obj) {
  assert(obj != nullptr);
  return obj->sockfd;
}

void CONN::client_open(const char* port_, const char* hostname_) {
  assert(port_ != nullptr);
  assert(hostname_ != nullptr);
  printf("II %s: port=%s hostname=%s\n", FUNCTION, port_, hostname_);

  FD_ZERO(&fds);
  typ = CONN_CLIENT;

  strcpy_s(port, sizeof port, port_);
  strcpy_s(hostname, sizeof hostname, hostname_);
  client_open1();
}

void CONN::client_open1() {
  struct addrinfo hints;
  memset(&hints, 0, sizeof hints);
  hints.ai_family = AF_UNSPEC;
  hints.ai_socktype = SOCK_STREAM;

  struct addrinfo* res;
  int rv = getaddrinfo(hostname, port, &hints, &res);
  if (rv != 0) {
    fprintf(stderr, "EE %s getaddrinfo: %d\n", FUNCTION, rv);
    exit(EXIT_FAILURE);
  }

  // loop through all the results and connect to the first we can
  struct addrinfo* p;
  for (p = res; p != nullptr; p = p->ai_next) {
    sockfd = socket(p->ai_family, p->ai_socktype, p->ai_protocol);
    if (INVALID_SOCKET == sockfd) {
      perror("WW socket");
      continue;
    }

    // UNPv2 ch. 15.4, 16.2 non-blocking connect
    unsigned long val = 1;
    rv = ioctlsocket(sockfd, FIONBIO, &val);
    if (rv != 0) {
      perror("WW ioctlsocket FIONBIO ON");
      closesocket(sockfd);
      continue;
    }

    rv = connect(sockfd, p->ai_addr, p->ai_addrlen);
    if (rv != 0) {
      if (WSAEWOULDBLOCK == WSAGetLastError()) {
        isConnecting = 1;
      } else {
        perror("WW connect");
        closesocket(sockfd);
        continue;
      }
    }

    break;  // exit loop after socket and connect were successful
  }

  assert(p != nullptr && "connect try");

  void* src = get_in_addr(reinterpret_cast<struct sockaddr*>(p->ai_addr));
  char dst[INET6_ADDRSTRLEN];
  inet_ntop(p->ai_family, src, dst, sizeof dst);

  freeaddrinfo(res);

  FD_SET(sockfd, &fds);

  printf("II %s: connect try to %s (%s) port %s socket %llu\n", FUNCTION,
         hostname, dst, port, sockfd);
}

void CONN::client_reopen() {
  closesocket(sockfd);
  FD_CLR(sockfd, &fds);   // remove network fd
  sockfd = -1;
  client_open1();
}

void CONN::client_handle(SOCKET fd) {
  assert(fd < FDMAX);
  int rv = cb_client_read(fd);
  if (rv < 1) {
    // documentation conflict between
    // https://docs.microsoft.com/en-us/windows/win32/api/winsock/nf-winsock-getsockopt
    // https://docs.microsoft.com/en-us/windows/win32/winsock/sol-socket-socket-options
    unsigned long optval;
    int optlen = sizeof optval;
    int rv = getsockopt(sockfd, SOL_SOCKET, SO_ERROR, 
                        reinterpret_cast<char*>(&optval), &optlen);
    assert(0 == rv && "getsockopt SOL_SOCKET SO_ERROR");
    fprintf(stderr, "WW %s: connect fail to %s port %s socket %llu rv %d: %s\n",
            FUNCTION, hostname, port, sockfd, rv, strerror(optval));
    client_reopen();
  } else {
    isConnecting = 0;   // hack: connect successful after first good read()
  }
}

void CONN::server_open(const char* port, const char* hostname) {
  assert(port != nullptr);
  assert(hostname != nullptr);
  printf("II %s: port=%s hostname=%s\n", FUNCTION, port, hostname);

  FD_ZERO(&fds);
  typ = CONN_SERVER;

  struct addrinfo hints;
  memset(&hints, 0, sizeof hints);
  hints.ai_family = AF_UNSPEC;
  hints.ai_socktype = SOCK_STREAM;
  hints.ai_flags = AI_PASSIVE;

  struct addrinfo* res;
  int rv = getaddrinfo(hostname, port, &hints, &res);
  if (rv != 0) {
    fprintf(stderr, "EE %s getaddrinfo: %d\n", FUNCTION, rv);
    exit(EXIT_FAILURE);
  }

  struct addrinfo* p;
  for(p = res; p != nullptr; p = p->ai_next) {
    sockfd = socket(p->ai_family, p->ai_socktype, p->ai_protocol);
    if (INVALID_SOCKET == sockfd) {
      perror("WW socket");
      continue;
    }

    // documentation conflict between
    // https://docs.microsoft.com/en-us/windows/win32/api/winsock/nf-winsock-setsockopt
    // https://docs.microsoft.com/en-us/windows/win32/winsock/sol-socket-socket-options
    const unsigned long yes = 1;
    rv = setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR,
                    reinterpret_cast<const char*>(&yes), sizeof(yes));
    if (rv != 0) {
      perror("WW setsockopt SO_REUSEADDR");
      closesocket(sockfd);
      continue;
    }

    rv = bind(sockfd, p->ai_addr, p->ai_addrlen);
    if (rv != 0) {
      perror("WW bind");
      closesocket(sockfd);
      continue;
    }

    break;  // exit loop after socket and bind were successful
  }

  freeaddrinfo(res);

  assert(p != NULL && "bind");

  rv = listen(sockfd, 10);
  assert(0 == rv && "listen");

  // add the listener to the master set
  FD_SET(sockfd, &fds);

  printf("II %s: listen on socket %llu\n", FUNCTION, sockfd);
}

void CONN::server_handle(SOCKET fd) {
  assert(fd < FDMAX);
  if (fd == sockfd) {
    // handle new connections
    struct sockaddr_storage remoteaddr; // client address
    socklen_t addrlen = sizeof remoteaddr;
    // newly accept()ed socket descriptor
    SOCKET newfd = accept(sockfd, reinterpret_cast<struct sockaddr*>(&remoteaddr),
                       &addrlen);
    if (INVALID_SOCKET == newfd) {
      perror("WW accept");
    } else {
      FD_SET(newfd, &fds); // add to master set
      void* src = get_in_addr(reinterpret_cast<struct sockaddr*>(&remoteaddr));
      char dst[INET6_ADDRSTRLEN];
      inet_ntop(remoteaddr.ss_family, src, dst, sizeof dst);
      printf("II %s: new connection %s on socket %llu\n", FUNCTION, dst, newfd);
    }
  } else {
    int rv = cb_server_read(fd);
    if (rv < 1 || SOCKET_ERROR == rv) {
      printf("II %s: connection closed on socket %llu\n", FUNCTION, fd);
      closesocket(fd);
      FD_CLR(fd, &fds); // remove from fd_set
    }
  }
}

void CONN::event_loop() {
  for(;;) {
    // virtualization pattern: join all read fds into one
    fd_set read_fds = conn[0].fds;
    for (int iconn = 1; iconn < CONNMAX; ++iconn) {
      for (SOCKET fd = 0; fd < FDMAX; ++fd) {
        if (FD_ISSET(fd, &conn[iconn].fds)) {
          FD_SET(fd, &read_fds);
        }
      }
    }
    // virtualization pattern: join all connect pending into one
    fd_set except_fds;
    FD_ZERO(&except_fds);
    for (int iconn = 0; iconn < CONNMAX; ++iconn) {
      if (conn[iconn].isConnecting) {
        FD_SET(conn[iconn].sockfd, &except_fds);
      }
    }

    struct timeval tv = {0, TIMEOUT * 1000};
    int rv = select(FDMAX, &read_fds, nullptr, &except_fds, &tv);
    if (SOCKET_ERROR == rv && WSAGetLastError() != WSAEINTR) {
      perror("EE select");
      exit(EXIT_FAILURE);
    }

    if (rv > 0) {
      // looking for data to read available
      for (SOCKET fd = 0; fd < FDMAX; ++fd) {
        if (FD_ISSET(fd, &read_fds)) {
          for (int iconn = 0; iconn < CONNMAX; ++iconn) {
            if (FD_ISSET(fd, &conn[iconn].fds)) {
              switch (conn[iconn].typ) {
              case CONN_CLIENT:
                conn[iconn].client_handle(fd);
                break;
              case CONN_SERVER:
                conn[iconn].server_handle(fd);
                break;
              }
            }
          }
        }
      }
      // looking for connect pending fail
      for (int iconn = 0; iconn < CONNMAX; ++iconn) {
        if (FD_ISSET(conn[iconn].sockfd, &except_fds)) {
          conn[iconn].client_reopen();
        }
      }
    }
    TIMER::walk();
  }
}

/* ******************************************************** */
// Business logic

/** @brief callback client read available
 * @retval >0 okay
 * @retval <=0 close or error
 */
int CONN::cb_client_read(SOCKET fd) {
  assert(fd < FDMAX);
  char buf[BUFSIZE];
  int readbytes = recv(fd, buf, sizeof buf, 0);
  if (readbytes > 0 && readbytes != SOCKET_ERROR) {
    // Write all data out
    int writebytes = fwrite(buf, 1, readbytes, stdout);
    assert(writebytes == readbytes && "fwrite");
  }
  return readbytes;
}

/** @brief callback server read available
 * @retval >0 okay
 * @retval <=0 close or error
 */
int CONN::cb_server_read(SOCKET fd) {
  assert(fd < FDMAX);
  char buf[BUFSIZE];    // buffer for client data
  int readbytes = recv(fd, buf, sizeof buf, 0);
  if (readbytes > 0 && readbytes != SOCKET_ERROR) {
    // we got some data from a client
    for (SOCKET i = 0; i < FDMAX; ++i) {
      // send to everyone!
      if (FD_ISSET(i, &fds)) {
        // except the listener and ourselves
        if (i != sockfd && i != fd) {
          int writebytes = send(i, buf, readbytes, 0);
          if (writebytes != readbytes) {
            perror("WW send");
          }
        }
      }
    }
  }
  return readbytes;
}

void poll_keyboard(CONN* obj) {
  assert(obj >= &conn[0] && obj < &conn[CONNMAX]);
  if (_kbhit()) {   // very MS-DOS
    char buf[BUFSIZE];
    char* rv = fgets(buf, sizeof buf, stdin);
    assert(rv != NULL);
    send(CONN::get_sockfd(obj), buf, strlen(buf), 0);
  }
  TIMER::after(TIMEOUT, (cb_timer_t)poll_keyboard, obj);
}

int main(int argc, char* argv[]) {
  WSADATA wsaData;
  int rv = WSAStartup(MAKEWORD(2, 2), &wsaData);
  assert(0 == rv && "WSAStartup");

  if (argc < 3) {
    char* name = basename(argv[0]);
    fprintf(stderr,"usage server: %s s port hostname\n", name);
    fprintf(stderr,"usage client: %s c hostname port\n", name);
    fprintf(stderr,"example server IPv4: %s s 60000 127.0.0.1\n", name);
    fprintf(stderr,"example client IPv4: %s c 127.0.0.1 60000\n", name);
    fprintf(stderr,"example server IPv6: %s s 60000 ::1\n", name);
    fprintf(stderr,"example client IPv6: %s c ::1 60000\n", name);
    exit(EXIT_FAILURE);
  }

  switch(argv[1][0]) {
  case 'c': {
    CONN* obj = CONN::factory();
    obj->client_open(argv[3], argv[2]);
    TIMER::after(TIMEOUT, reinterpret_cast<cb_timer_t>(poll_keyboard), obj);
  }
  break;
  case 's': {
    CONN* obj = CONN::factory();
    obj->server_open(argv[2], argv[3]);
  }
  break;
  default:
    fprintf(stderr,"EE %s: unexpected argument %s\n", FUNCTION, argv[1]);
    exit(EXIT_FAILURE);
  }

  CONN::event_loop();  // start inversion of control

  WSACleanup();
  return 0;
}
